from itertools import count
from typing import Iterator
from typing import List
from typing import Union

from pydantic import parse_obj_as
from requests import Response

from _incydr_sdk.devices.models import DevicesPage
from _incydr_sdk.devices.models import QueryDevicesRequest
from _incydr_sdk.enums import SortDirection
from _incydr_sdk.enums.devices import SortKeys
from _incydr_sdk.exceptions import IncydrException
from _incydr_sdk.users.models import QueryUsersRequest
from _incydr_sdk.users.models import Role
from _incydr_sdk.users.models import UpdateRolesResponse
from _incydr_sdk.users.models import User
from _incydr_sdk.users.models import UserRole
from _incydr_sdk.users.models import UsersPage


class UsersClient:
    def __init__(self, parent):
        self._parent = parent
        self._v1 = None

    @property
    def v1(self):
        if self._v1 is None:
            self._v1 = UsersV1(self._parent)
        return self._v1


class UsersV1:
    """Client for `/v1/users` endpoints.

    Usage example:

        >>> import incydr
        >>> client = incydr.Client(**kwargs)
        >>> client.users.v1.get_page()
    """

    def __init__(self, parent):
        self._parent = parent
        self._available_roles = {}

    def get_user(self, user: str) -> User:
        """
        Get a single user.

        **Parameters:**

        * **user**: `str` - The unique ID for the user or the username for the user.

        **Returns**: A [`User`][user-model] object representing the user.
        """

        # if username, lookup with get page api
        if "@" in user:
            users = self.get_page(username=user).users
            if len(users) < 1:
                raise ValueError(f"User with username '{user}' not found.")
            return users[0]

        # if user ID, use GET api
        response = self._parent.session.get(f"/v1/users/{user}")
        return User.parse_response(response)

    def get_page(
        self,
        active: bool = None,
        blocked: bool = None,
        username: str = None,
        page_num: int = 1,
        page_size: int = None,
    ) -> UsersPage:
        """
        Get a page of users.

        Filter results by passing the appropriate parameters:

        **Parameters**:

        * **active**: `bool | None` - When true, return only active users. When false, return only inactive users. Defaults to returning both.
        * **blocked**: `bool | None` - When true, return only blocked users. When false, return only unblocked users. Defaults to returning both.
        * **username**: `str` - The username of a user to search for.
        * **page_num**: `int` - Page number for results. Defaulting to 1.
        * **page_size**: `int` - Max number of results to return per page. Defaulting to client's `page_size` setting.

        **Returns**: A [`UsersPage`][userspage-model] object.
        """
        page_size = page_size or self._parent.settings.page_size
        data = QueryUsersRequest(
            active=active,
            blocked=blocked,
            username=username,
            page=page_num,
            pageSize=page_size,
        )
        response = self._parent.session.get("/v1/users", params=data.dict())
        return UsersPage.parse_response(response)

    def get_devices(
        self,
        user_id: str,
        active: bool = None,
        blocked: bool = None,
        page_num: int = 1,
        page_size: int = None,
        sort_dir: SortDirection = SortDirection.ASC,
        sort_key: SortKeys = SortKeys.NAME,
    ) -> User:
        """
        Get a page of devices associated with a specific user.

        Filter results by passing the appropriate parameters:

        **Parameters**:

        * **user_id**: `str` (required) - The unique ID for the user.
        * **active**: `bool` - Whether or not the device is active. If true, the device will show up on reports, etc.
        * **blocked**: `bool` - Whether or not the device is blocked.  If true, restores and logins are disabled.
        * **page_num**: `int` - Page number for results. Defaulting to 1.
        * **page_size**: `int` - Max number of results to return per page. Defaulting to client's `page_size` settings.
        * **sort_dir**: `SortDirection` - 'asc' or 'desc'. The direction in which to sort the response based on the corresponding key. Defaults to 'asc'.
        * **sort_key**: `SortKeys` - One or more values on which the response will be sorted. Defaults to device name.

        **Returns**: A [`DevicesPage`][devicespage-model] object.
        """
        page_size = page_size or self._parent.settings.page_size
        data = QueryDevicesRequest(
            page=page_num,
            pageSize=page_size,
            sortKey=sort_key,
            sortDirection=sort_dir,
            active=active,
            blocked=blocked,
        )
        response = self._parent.session.get(
            f"/v1/users/{user_id}/devices", params=data.dict()
        )
        return DevicesPage.parse_response(response)

    def iter_all(
        self,
        active: bool = None,
        blocked: bool = None,
        username: str = None,
        page_size: int = None,
    ) -> Iterator[User]:
        """
        Iterate over all users.

        Accepts the same parameters as `.get_page()` excepting `page_num`.

        **Returns**: A generator yielding individual [`User`][user-model] objects.
        """
        page_size = page_size or self._parent.settings.page_size
        for page_num in count(1):
            page = self.get_page(
                active=active,
                blocked=blocked,
                username=username,
                page_num=page_num,
                page_size=page_size,
            )
            yield from page.users
            if len(page.users) < page_size:
                break

    def list_user_roles(
        self,
        user_id: str,
    ) -> List[UserRole]:
        """
        Get a list of roles associated with a specific user.

        **Parameters**:

        * **user_id**: `str` (required) - The unique ID for the user.

        **Returns**: A list of [`UserRole`][role-model] objects.
        """
        response = self._parent.session.get(f"/v1/users/{user_id}/roles")
        return parse_obj_as(List[UserRole], response.json()["roles"])

    def update_roles(
        self, user_id: str, roles: Union[str, List[str]]
    ) -> UpdateRolesResponse:
        """
        Replace the roles associated with a user.

        **Parameters**:

        * **user_id**: `str` (required) - The unique ID for the user.
        * **roles**: `str | List[str]` The new roles to assign the user (ex: desktop-user). These will replace the existing roles assigned to the user. Accepts either role IDs or role names."

        **Returns**: A [`UpdateRolesResponse`][updaterolesresponse-model] object.
        """
        if not isinstance(roles, List):
            roles = [roles]

        roles = [self._get_id_by_name(role) for role in roles]

        response = self._parent.session.put(
            f"/v1/users/{user_id}/roles", json={"roleIds": roles}
        )
        return UpdateRolesResponse.parse_response(response)

    def add_roles(
        self, user_id: str, roles: Union[str, List[str]]
    ) -> UpdateRolesResponse:
        """
        Add a role, or multiple roles, to a user's existing roles.

        **Parameters**:

        * **user_id**: `str` (required) - The unique ID for the user.
        * **roles**: `str | List[str]` The roles to add to the user. Accepts either role IDs or role names (case-sensitive)."

        **Returns**: A [`UpdateRolesResponse`][updaterolesresponse-model] object.
        """
        roles = self._update_role_ids_for_user(roles, user_id, add=True)
        response = self._parent.session.put(
            f"/v1/users/{user_id}/roles", json={"roleIds": roles}
        )
        return UpdateRolesResponse.parse_response(response)

    def remove_roles(
        self, user_id: str, roles: Union[str, List[str]]
    ) -> UpdateRolesResponse:
        """
        Remove a role, or multiple roles, from a user's current roles.

        **Parameters**:

        * **user_id**: `str` (required) - The unique ID for the user.
        * **roles**: `str | List[str]` The roles to remove from the user. Accepts either role IDs or role names."

        **Returns**: A `requests.Response` indicating success.
        """
        roles = self._update_role_ids_for_user(roles, user_id, add=False)
        response = self._parent.session.put(
            f"/v1/users/{user_id}/roles", json={"roleIds": roles}
        )
        return UpdateRolesResponse.parse_response(response)

    def move(self, user_id: str, org_guid: str) -> Response:
        """
        Move a user to a specified organization

        **Parameters**:

        * **user_id**: `str` - The unique ID for the user.
        * **org_guid**: `str` The orgGuid of the org to move the user to."

        **Returns**: A `requests.Response` indicating success.
        """
        return self._parent.session.post(
            f"/v1/users/{user_id}/move", json={"orgGuid": org_guid}
        )

    def activate(self, user_id: str) -> Response:
        """
        Activate a user.

        **Parameters:**

        * **user_id**: `str` (required) - The unique ID for the user.

        **Returns**: A `requests.Response` indicating success.
        """
        return self._parent.session.post(f"/v1/users/{user_id}/activate")

    def deactivate(self, user_id: str) -> Response:
        """
        Deactivate a user.

        **Parameters:**

        * **user_id**: `str` (required) - The unique ID for the user.

        **Returns**: A `requests.Response` indicating success.
        """
        return self._parent.session.post(f"/v1/users/{user_id}/deactivate")

    def list_roles(self) -> List[Role]:
        """
        Get a list of all available roles that can be assigned by the current user.

        **Parameters**:

        **Returns**: A list of [`Role`][role-model] objects.
        """
        response = self._parent.session.get("/v1/users/roles")
        return parse_obj_as(List[Role], response.json())

    def get_role(self, role: str) -> Role:
        """
        Get details for a single role.

        **Parameters**:

        * **role**: `str` (required) - Role ID or role name (case-sensitive).

        **Returns**: A [`Role`](role-model) object.
        """
        response = self._parent.session.get(
            f"/v1/users/roles/{self._get_id_by_name(role)}"
        )
        return Role.parse_response(response)

    def _get_id_by_name(self, role_name: str):
        """
        Get a role ID by its name.

        Returns the role ID unchanged if it doesn't match any names of available roles.
        """
        if not self._available_roles:
            self._lookup_roles()
        for name, id_ in self._available_roles.items():
            if (role_name == name) or (role_name == id_):
                return id_
        raise RoleNotFoundError(role_name)

    def _update_role_ids_for_user(self, roles, user_id, add=True):
        """
        Adds or removes a role ID specified by `role` to the list of user's role IDs,
        which can be passed to the update roles method.

        Where `role_name` is either a role name or a role ID.

        Returns the updated list of role IDs for a user.
        """
        errors = []

        role_ids = [i.role_id for i in self.list_user_roles(user_id)]

        if not self._available_roles:
            self._lookup_roles()

        if not isinstance(roles, List):
            roles = [roles]

        for role in roles:
            try:
                id_ = self._get_id_by_name(role)
            except RoleNotFoundError as err:
                errors.append(err)
                continue
            if add:
                role_ids.append(id_)
            else:
                try:
                    role_ids.remove(id_)
                except ValueError:
                    errors.append(UserNotAssignedRoleError(id_))

        if errors:
            raise RoleProcessingError(errors)

        return role_ids

    def _lookup_roles(self):
        """Map role names to role ID."""
        self._available_roles = {}
        available_roles = self.list_roles()
        for r in available_roles:
            self._available_roles[r.role_name] = r.role_id


class RoleProcessingError(IncydrException):
    """
    Outputs list of errors that arose during processing.

    Example output:
        incydr._users.client.RoleProcessingError: The following errors arose during role processing:
            * User is not currently assigned the following role: 'alert-emails'. Role cannot be removed.
            * No role matching the following was found: 'fake', or you do not have permission to assign this role.
    """

    def __init__(self, errors):
        message = (
            "The following errors arose during role processing:\n\t* "
            + "\n\t* ".join([str(e) for e in errors])
        )
        super().__init__(message)
        self._errors = errors

    @property
    def errors(self):
        """List of errors that arose during role processing."""
        return self._errors


class UserNotAssignedRoleError(IncydrException):
    def __init__(self, role):
        message = f"User Not Assigned Role Error: User is not currently assigned the following role: '{role}'. Role cannot be removed."
        super().__init__(message)
        self._role = role

    @property
    def role(self):
        """The role which cannot be assigned."""
        return self._role


class RoleNotFoundError(IncydrException):
    def __init__(self, role):
        message = f"Role Not Found Error: No role matching the following was found: '{role}', or you do not have permission to assign this role. Roles can be specified by case-sensitive name (ie. 'Cloud Admin') or ID (ie. cloud-admin)."
        super().__init__(message)
        self._role = role

    @property
    def role(self):
        """The role which could not be found"""
        return self._role



from datetime import datetime
from itertools import count
from typing import Iterator
from typing import Union
from warnings import warn

from _incydr_sdk.exceptions import DateParseError
from _incydr_sdk.queries.utils import DATE_STR_FORMAT
from _incydr_sdk.risk_profiles.models import Date
from _incydr_sdk.risk_profiles.models import QueryRiskProfilesRequest
from _incydr_sdk.risk_profiles.models import RiskProfile
from _incydr_sdk.risk_profiles.models import RiskProfilesPage
from _incydr_sdk.risk_profiles.models import UpdateRiskProfileRequest


class RiskProfiles:
    def __init__(self, parent):
        self._parent = parent
        self._v1 = None

    @property
    def v1(self):
        if self._v1 is None:
            self._v1 = RiskProfilesV1(self._parent)
        return self._v1


class RiskProfilesV1:
    """
    Client for `/v1/user-risk-profiles` endpoints.

    This client is deprecated. Use the Actors client instead.

    Usage example:

        >>> import incydr
        >>> client = incydr.Client(**kwargs)
        >>> client.risk_profiles.v1.get_page()
    """

    def __init__(self, parent):
        self._parent = parent

    def get_risk_profile(self, user_id: str) -> RiskProfile:
        """
        Get a single risk profile.

        **Parameters:**

        * **user_id**: `str` (required) - The unique ID for the user.

        **Returns**: A [`RiskProfile`][riskprofile-model] object representing the risk profile.
        """
        warn(
            "Risk Profiles are deprecated. Replaced by Actors.",
            DeprecationWarning,
            stacklevel=2,
        )
        response = self._parent.session.get(f"/v1/user-risk-profiles/{user_id}")
        return RiskProfile.parse_response(response)

    def get_page(
        self,
        page_num: int = 1,
        page_size: int = None,
        manager_id: str = None,
        title: str = None,
        division: str = None,
        department: str = None,
        employment_type: str = None,
        country: str = None,
        region: str = None,
        locality: str = None,
        active: bool = None,
        deleted: bool = None,
        support_user: bool = None,
    ) -> RiskProfilesPage:
        """
        Get a page of risk profiles.

        Filter results by passing the appropriate parameters:

        **Parameters**:

        * **page_num**: `int` - Page number for results, starting at 1.
        * **page_size**: `int` - Max number of results to return per page.
        * **manager_id**: `str` - The Code42 user ID of the user's manager to search for.
        * **title**: `str` - The user's job title to search for.
        * **division**: `str` - The user's division to search for.
        * **department**: `str` - The user's department to search for.
        * **employment_type**: `str` - The user's employment type to search for.
        * **country**: `str` - The user's country to search for.
        * **region**: `str` - The user's region (state) to search for.
        * **locality**: `str` - The user's locality (city) to search for.
        ** **active**: `bool | None` - When true, return only active users. When false, return only inactive users.
                                       Defaults to `None` (returning both).
        * **deleted**: `bool | None` - When true, return only deleted users. When false, return only non-deleted users.
                                       Defaults to returning both.
        * **support_user**: `bool | None` - When true, return only support users. When false, return only non-support users.
                                            Defaults to returning both

        **Returns**: A ['RiskProfilesPage'][riskprofilespage-model] object.
        """
        warn(
            "Risk Profiles are deprecated. Replaced by Actors.",
            DeprecationWarning,
            stacklevel=2,
        )
        page_size = page_size or self._parent.settings.page_size
        data = QueryRiskProfilesRequest(
            page=page_num,
            page_size=page_size,
            manager_id=manager_id,
            title=title,
            division=division,
            department=department,
            employment_type=employment_type,
            country=country,
            region=region,
            locality=locality,
            active=active,
            deleted=deleted,
            support_user=support_user,
        )
        response = self._parent.session.get(
            "/v1/user-risk-profiles", params=data.dict()
        )
        return RiskProfilesPage.parse_response(response)

    def iter_all(
        self,
        page_size: int = None,
        manager_id: str = None,
        title: str = None,
        division: str = None,
        department: str = None,
        employment_type: str = None,
        country: str = None,
        region: str = None,
        locality: str = None,
        active: bool = None,
        deleted: bool = None,
        support_user: bool = None,
    ) -> Iterator[RiskProfile]:
        """
        Iterate over all risk profiles.

        Accepts the same parameters as `.get_page()` except `page_num`.

        **Returns**: A generator yielding individual [`RiskProfile`][riskprofile-model] objects.
        """
        warn(
            "Risk Profiles are deprecated. Replaced by Actors.",
            DeprecationWarning,
            stacklevel=2,
        )
        page_size = page_size or self._parent.settings.page_size
        for page_num in count(1):
            page = self.get_page(
                page_num=page_num,
                page_size=page_size,
                manager_id=manager_id,
                title=title,
                division=division,
                department=department,
                employment_type=employment_type,
                country=country,
                region=region,
                locality=locality,
                active=active,
                deleted=deleted,
                support_user=support_user,
            )
            yield from page.user_risk_profiles
            if len(page.user_risk_profiles) < page_size:
                break

    def update(
        self,
        user_id: str,
        notes: str = None,
        start_date: Union[datetime, str] = None,
        end_date: Union[datetime, str] = None,
    ) -> RiskProfile:
        """
        Updates a risk profile.

        **Parameters**

        * **notes**: `str` - Additional notes for the risk profile.
        * **start_date**: `datetime` - The starting date for the user. Accepts a datetime object or a string in the format yyyy-MM-dd (UTC) format. Pass an empty string to clear the field.
        * **end_date**: `datetime` - The departure date for the user.  Accepts a datetime object or a string in the format yyyy-MM-dd (UTC) format.  Pass an empty string to clear the field.

        **Returns**: A [`RiskProfile`][riskprofile-model] object.
        """
        warn(
            "Risk Profiles are deprecated. Replaced by Actors.",
            DeprecationWarning,
            stacklevel=2,
        )
        paths = []
        if start_date is not None:
            paths += ["startDate"]
            start_date = None if start_date == "" else _create_date(start_date)
        if end_date is not None:
            paths += ["endDate"]
            end_date = None if end_date == "" else _create_date(end_date)
        if notes is not None:
            paths += ["notes"]
            if notes == "":
                notes = None

        data = UpdateRiskProfileRequest(
            endDate=end_date, notes=notes, startDate=start_date
        )

        response = self._parent.session.patch(
            f"/v1/user-risk-profiles/{user_id}",
            params={"paths": paths},
            json=data.dict(),
        )
        return RiskProfile.parse_response(response)


def _create_date(date: Union[datetime, str]):
    if not isinstance(date, datetime):
        try:
            date = datetime.strptime(date, DATE_STR_FORMAT)
        except ValueError:
            raise DateParseError(
                date,
                msg=f"DateParseError: Error parsing time data. Date '{date}' does not match format {DATE_STR_FORMAT}.",
            )
    return Date(day=date.day, month=date.month, year=date.year)



import enum as _enum


class _Enum(str, _enum.Enum):
    """
    An `enum.Enum` subclass that enables string comparison (`Enum.MEMBER == "MEMBER"`) and better exceptions that show
    all possible values.
    """

    @classmethod
    def _missing_(cls, value):
        if value in cls.__members__:
            return None
        raise ValueError(
            f"'{value}' is not a valid {cls.__name__}. Expected one of {[member.value for member in cls]}"
        )


class SortDirection(_Enum):
    ASC = "asc"
    DESC = "desc"



from _incydr_sdk.enums import _Enum


class SessionStates(_Enum):
    """
    Enum indicating possible session states (includes alerts).
    """

    OPEN = "OPEN"
    IN_PROGRESS = "IN_PROGRESS"
    CLOSED = "CLOSED"
    CLOSED_TP = "CLOSED_TP"
    CLOSED_FP = "CLOSED_FP"
    OPEN_NEW_DATA = "OPEN_NEW_DATA"


class SessionSeverities(_Enum):
    """
    Enum indicating possible session severities.
    """

    NO_RISK = "NO_RISK"
    LOW = "LOW"
    MODERATE = "MODERATE"
    HIGH = "HIGH"
    CRITICAL = "CRITICAL"


class ContentInspectionStatuses(_Enum):
    """
    Enum indicating possible content inspection statuses.
    """

    PENDING = "PENDING"
    FOUND = "FOUND"
    NOT_FOUND = "NOT_FOUND"


class SortKeys(_Enum):
    """
    Enum indicating possible fields by which to sort items results.
    """

    END_TIME = "end_time"
    SCORE = "score"



import os
from functools import lru_cache
from pathlib import Path
from typing import Optional

import click
from pydantic import Field
from pydantic import root_validator
from requests.exceptions import HTTPError
from rich.panel import Panel
from rich.progress import track

from _incydr_cli import console
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import input_format_option
from _incydr_cli.cmds.options.output_options import single_format_option
from _incydr_cli.cmds.options.output_options import SingleFormat
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.cmds.options.utils import user_lookup_callback
from _incydr_cli.cmds.utils import user_lookup
from _incydr_cli.core import incompatible_with
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_cli.file_readers import FileOrString
from _incydr_sdk.cases.models import Case
from _incydr_sdk.cases.models import CaseDetail
from _incydr_sdk.cases.models import FileEvent
from _incydr_sdk.core.client import Client
from _incydr_sdk.core.models import CSVModel
from _incydr_sdk.file_events.models.event import FileEventV2
from _incydr_sdk.utils import model_as_card


class UpdateCaseCSV(CSVModel):
    number: str = Field(csv_aliases=["number", "case_number"])
    assignee: Optional[str]
    description: Optional[str]
    findings: Optional[str]
    name: Optional[str]
    status: Optional[str]
    subject: Optional[str]


class FileEventCSV(CSVModel):
    event_id: str = Field(csv_aliases=["event_id", "eventId"])


class FileEventJSON(FileEventV2):
    event_id: str = Field(alias="eventId")

    @root_validator(pre=True)
    def _event_id_required(cls, values):  # noqa
        # check if input is V2 file event
        event = values.get("event")
        if event and event.get("id"):
            values["event_id"] = event["id"]
        return values


path_option = click.option(
    "--path",
    help="The file path where to save the file. Defaults to the current directory.",
    default=os.getcwd(),
)


@click.group(cls=IncydrGroup)
@logging_options
def cases():
    """View and manage cases."""


@cases.command(cls=IncydrCommand)
@click.argument("name")
@click.option("--description", help="Case description.", default=None)
@click.option(
    "--subject",
    help="User of the subject of the case.  Takes a user ID or a username.  Performs an additional lookup if a username is passed.",
    default=None,
    callback=user_lookup_callback,
)
@click.option(
    "--assignee",
    help="User of the assignee of the case. Takes an actor ID or a username.  Performs an additional lookup if a username is passed.",
    default=None,
    callback=user_lookup_callback,
)
@click.option(
    "--findings", help="Markdown formatted details of case notes.", default=None
)
@logging_options
def create(
    name: str,
    description: str,
    subject: str,
    assignee: str,
    findings: str,
):
    """
    Create a case.
    """
    client = Client()
    try:
        case = client.cases.v1.create(
            name,
            subject=subject,
            assignee=assignee,
            findings=findings,
            description=description,
        )
        if client.settings.use_rich:
            console.print(Panel.fit(model_as_card(case), title="Case"))
        else:
            console.print(case.json(), highlight=False)
    except HTTPError as err:
        console.print(f"[red]Error:[/red] {err.response.text}")


@cases.command(cls=IncydrCommand)
@click.argument("case_number")
@logging_options
def delete(case_number: int):
    """
    Delete a case.
    """
    client = Client()
    client.cases.v1.delete(case_number)
    console.print(f"Case number '{case_number}' successfully deleted.")


@cases.command("list", cls=IncydrCommand)
@table_format_option
@columns_option
@logging_options
def list_(
    format_: TableFormat,
    columns: Optional[str],
):
    """
    List all cases.
    """
    client = Client()
    cases_ = client.cases.v1.iter_all()

    if format_ == TableFormat.table:
        render.table(Case, cases_, columns=columns)
    elif format_ == TableFormat.csv:
        render.csv(Case, cases_, columns=columns)
    elif format_ == TableFormat.json_pretty:
        for case in cases_:
            console.print_json(case.json())
    else:
        for case in cases_:
            click.echo(case.json())


@cases.command(cls=IncydrCommand)
@click.argument("case_number")
@single_format_option
@logging_options
def show(
    case_number: int,
    format_: SingleFormat,
):
    """
    Show details for a single case.
    """
    client = Client()
    case = client.cases.v1.get_case(case_number)
    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(Panel.fit(model_as_card(case), title="Case", width=120))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(case.json())
    else:
        click.echo(case.json())


@cases.command(cls=IncydrCommand)
@click.argument("case_number")
@click.option(
    "--assignee",
    default=None,
    help="The administrator assigned to the case. Takes a user ID or a username.  Performs an additional lookup if a username is passed.",
    callback=user_lookup_callback,
)
@click.option("--description", default=None, help="Brief optional description.")
@click.option(
    "--findings",
    default=None,
    help="Markdown formatted text summarizing the findings for a case.",
)
@click.option("--name", default=None, help="Case name.")
@click.option(
    "--status", default=None, help="Case status. One of `ARCHIVED`, `CLOSED` or `OPEN`."
)
@click.option(
    "--subject",
    default=None,
    help="The case subject. Takes a user ID or a username.  Performs an additional lookup if a username is passed.",
    callback=user_lookup_callback,
)
@logging_options
def update(
    case_number: str,
    assignee: Optional[str] = None,
    description: Optional[str] = None,
    findings: Optional[str] = None,
    name: Optional[str] = None,
    status: Optional[str] = None,
    subject: Optional[str] = None,
):
    """
    Update a single case. Pass the updated value for a field to the corresponding command option.
    """
    if not any([assignee, description, findings, name, status, subject]):
        raise click.UsageError(
            "At least one command option must be provided to update a case.  Use `cases update --help` to see available options."
        )

    client = Client()
    case = client.cases.v1.get_case(case_number)
    if assignee:
        case.assignee = assignee
    if description:
        case.description = description
    if findings:
        case.findings = findings
    if name:
        case.name = name
    if status:
        case.status = status
    if subject:
        case.subject = subject
    updated_case = client.cases.v1.update(case)
    if client.settings.use_rich:
        console.print(Panel.fit(model_as_card(updated_case), title="Updated Case"))
    else:
        console.print(updated_case.json(), highlight=False)


@cases.command(cls=IncydrCommand)
@click.argument("file", type=click.File())
@input_format_option
@logging_options
def bulk_update(
    file: Path,
    format_: str,
):
    """
    Bulk update cases from a file.

    Takes a single arg `FILE` which specifies the path to the file (use "-" to read from stdin).

    File format can either be CSV or [JSON Lines format](https://jsonlines.org) (Default is CSV).

    Valid CSV columns that correspond to update-able case fields include:

    * `number` (REQUIRED) - Case number.
    * `assignee` - actor ID or username of the administrator assigned to the case. Performs an additional lookup if a username is passed.
    * `description` - Brief optional description.
    * `findings` - Markdown formatted text summarizing the findings for a case.
    * `name` - Case name.
    * `status` - Case status. One of `ARCHIVED`, `CLOSED` or `OPEN`.
    * `subject` - User ID or username of the case subject. Performs an additional lookup if a username is passed.
    """
    client = Client()

    @lru_cache()
    def resolve_username(user):
        if user is None:
            return
        elif "@" in user:
            return user_lookup(client, user)
        else:  # assume user_id
            return user

    if format_ == "csv":
        models = UpdateCaseCSV.parse_csv(file)
    else:
        models = CaseDetail.parse_json_lines(file)
    try:
        for updated in track(models, description="Updating cases...", transient=True):
            fields_set = updated.__fields_set__
            case = client.cases.v1.get_case(updated.number)
            if "assignee" in fields_set and case.assignee != updated.assignee:
                case.assignee = resolve_username(updated.assignee)
            if "subject" in fields_set and case.subject != updated.subject:
                case.subject = resolve_username(updated.subject)
            if "description" in fields_set:
                case.description = updated.description
            if "findings" in fields_set:
                case.findings = updated.findings
            if "name" in fields_set:
                case.name = updated.name
            if "status" in fields_set:
                case.status = updated.status
            client.cases.v1.update(case)
    except ValueError as err:
        console.print(f"[red]Error:[/red] {err}")
    except HTTPError as err:
        console.print(f"[red]Error:[/red] {err.response.text}")


@cases.command(cls=IncydrCommand)
@click.argument("case_number")
@path_option
@click.option(
    "--summary",
    is_flag=True,
    default=False,
    help="Download a case summary in PDF format.",
)
@click.option(
    "--file-events",
    "events",
    is_flag=True,
    default=False,
    help="Download all file event data for a case in CSV format.",
)
@click.option(
    "--source-files",
    is_flag=True,
    default=False,
    help="Download the source files for file events associated with a case.",
)
@click.option(
    "--source-file",
    default=None,
    help="Download a source file for a specific event. Takes the event ID. Incompatible with other download options.",
    cls=incompatible_with(["summary", "file_events", "source_files"]),
)
@logging_options
def download(
    case_number: int,
    path: str,
    summary: bool = None,
    events: bool = None,
    source_files: bool = None,
    source_file: str = None,
):
    """
    Download one or more files related to a case to the specified target folder.

    Defaults to downloading all files if no options are passed.

    If more than one file is specified the download will be in ZIP format.
    """
    client = Client()

    # download source file for specific event
    if source_file:
        client.cases.v1.download_file_for_event(case_number, source_file, path)
        return

    flags = [summary, events, source_files]

    # if only summary or only file-events flag is passed
    if sum(bool(flag) for flag in flags) == 1:
        if summary:
            client.cases.v1.download_summary_pdf(case_number, path)
            return
        if events:
            client.cases.v1.download_file_event_csv(case_number, path)
            return

    # Default to downloading all files if no flags are passed
    if not any(flags):
        summary = True
        events = True
        source_files = True

    client.cases.v1.download_full_case_zip(
        case_number,
        path,
        include_files=source_files,
        include_summary=summary,
        include_file_events=events,
    )
    console.print(f"Case file(s) downloaded to '{path}'")


@cases.group(cls=IncydrGroup)
def file_events():
    """View and update file events associated with a case."""


@file_events.command("show", cls=IncydrCommand)
@click.argument("case_number")
@click.argument("event_id")
@single_format_option
@logging_options
def show_file_event(
    case_number: int,
    event_id: str,
    format_: SingleFormat,
):
    """
    Show details for a file event attached to a case.
    """
    client = Client()
    event = client.cases.v1.get_file_event_detail(case_number, event_id)

    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(Panel.fit(model_as_card(event)))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(event.json())
    else:
        click.echo(event.json())


@file_events.command("list", cls=IncydrCommand)
@click.argument("case_number")
@table_format_option
@columns_option
@logging_options
def list_file_events(
    case_number: int,
    format_: TableFormat,
    columns: str,
):
    """
    List file events attached to a case.
    """
    client = Client()
    response = client.cases.v1.get_file_events(case_number)

    if format_ == TableFormat.table:
        columns = columns or [
            "event_timestamp",
            "file_name",
            "file_path",
            "file_availability",
            "risk_indicators",
            "risk_score",
            "event_id",
        ]
        render.table(
            FileEvent,
            response.events,
            title=f"Case {case_number}: File Events",
            columns=columns,
        )
    elif format_ == TableFormat.csv:
        render.csv(FileEvent, response.events, columns=columns)

    else:
        printed = False
        if format_ == TableFormat.json_pretty:
            for event in response.events:
                printed = True
                console.print_json(event.json())

        else:
            for event in response.events:
                printed = True
                click.echo(event.json())
        if not printed:
            console.print("No results found.")


csv_option = click.option(
    "--csv", is_flag=True, default=False, help="event IDs are specified in a CSV file."
)


@file_events.command(cls=IncydrCommand)
@click.argument("case_number")
@click.argument("event_ids", type=FileOrString())
@click.option(
    "--format", "-f", "format_", type=click.Choice(["csv", "json-lines"]), default="csv"
)
@logging_options
def add(
    case_number: int,
    event_ids: FileOrString,
    format_: str,
):
    """
    Attach file events to a case specified by CASE_NUMBER.

    EVENT_IDS can be either a comma-delimited string of event IDs:

        add CASE_NUMBER "id-1,id-2,id-3,..."

    Or a CSV or [JSON Lines](https://jsonlines.org) formatted file:

        add CASE_NUMBER @path_to_csv --format csv

        add CASE_NUMBER @path_to_json --format json-lines

    CSV format requires a header row and a column with name matching either "event_id" or "eventId".

    Input can also be parsed from stdin using "-" as the command argument, so you can add events
    directly from an `incydr file-events search` command to a case:

        incydr file-events search SEARCH_OPTIONS --format json-lines | incydr cases add CASE_NUMBER --format json-lines -

    """
    client = Client()
    if isinstance(event_ids, str):
        event_ids = [e.strip() for e in event_ids.split(",")]
    elif format_ == "csv":
        events = FileEventCSV.parse_csv(event_ids)
        event_ids = [e.event_id for e in events]
    else:
        events = FileEventJSON.parse_json_lines(event_ids)
        event_ids = [e.event_id for e in events]

    client.cases.v1.add_file_events_to_case(case_number, event_ids)
    console.print(f"{len(event_ids)} events added to case {case_number}")


@file_events.command(cls=IncydrCommand)
@click.argument("case_number")
@click.argument("event_ids", type=FileOrString())
@input_format_option
@click.option(
    "--format", "-f", "format_", type=click.Choice(["csv", "json-lines"]), default="csv"
)
@logging_options
def remove(
    case_number: int,
    event_ids: str,
    format_: str,
):
    """
    Remove file events from a case specified by CASE_NUMBER.

    EVENT_IDS can be either a comma-delimited string of event IDs:

        remove CASE_NUMBER "id-1,id-2,id-3,..."

    Or a CSV or [JSON Lines](https://jsonlines.org) formatted file:

        remove CASE_NUMBER @path_to_csv --format csv

        remove CASE_NUMBER @path_to_json --format json-lines

    CSV format requires a header row and a column with name matching either "event_id" or "eventId".

    Input can also be parsed from stdin using "-" as the command argument, so you can remove all events from a case
    by sending the output of `cases file-events list` command into `cases file-events remove`:

        incydr cases file-events list CASE_NUMBER --format json-lines | incydr cases remove CASE_NUMBER --format json-lines -

    """
    client = Client()
    if isinstance(event_ids, str):
        event_ids = [e.strip() for e in event_ids.split(",")]
    elif format_ == "csv":
        events = FileEventCSV.parse_csv(event_ids)
        event_ids = [e.event_id for e in events]
    else:
        events = FileEventJSON.parse_json_lines(event_ids)
        event_ids = [e.event_id for e in events]

    for id_ in event_ids:
        try:
            client.cases.v1.delete_file_event_from_case(case_number, id_)
            console.print(f"Event removed: {id_}", highlight=False)
        except HTTPError as err:
            if err.response.status_code == 404:
                console.print(f"Event not found on case: {id_}", highlight=False)
            else:
                console.print(
                    f"[red]Error removing event:[/red] {id_} ({err.response.status_code})",
                    highlight=False,
                )


if __name__ == "__main__":
    cases()



from __future__ import annotations

from typing import List
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from _incydr_sdk.alerts.models.alert import AlertDetails  # noqa
from _incydr_sdk.alerts.models.alert import AlertSummary
from _incydr_sdk.alerts.models.enums import ProblemType
from _incydr_sdk.core.models import ResponseModel


class QueryProblem(BaseModel):
    bad_filter: Optional[str] = Field(
        None, alias="badFilter", description="The filter that caused the issue."
    )
    type: ProblemType = Field(..., description="The type of query problem.")


class AlertQueryPage(ResponseModel):
    """
    A model representing a page of `AlertSummary` objects resulting from an alert search query.

    **Fields**:

    * **alerts**: `List[AlertSummary]` List of alerts that found by query.
    * **total_count**: `int` The count of alerts found.
    * **problems**: `List[QueryProblem]` Potential issues that were hit while trying to run the query.
    """

    type: str = Field(alias="type$")
    alerts: Optional[List[AlertSummary]] = Field(
        None, description="List of alerts that are returned."
    )
    total_count: int = Field(
        ...,
        alias="totalCount",
        description="The number of alerts that match the given query.",
        example="3",
    )
    problems: Optional[List[QueryProblem]] = Field(
        None,
        description="Potential issues that were hit while trying to run the query.",
        example=[],
    )



from _incydr_sdk.enums import _Enum


class SortKeys(_Enum):
    """Possible keys to sort cases list results by."""

    NAME = "name"
    NUMBER = "number"
    CREATED_AT = "createdAt"
    UPDATED_AT = "updatedAt"
    STATUS = "status"
    ASSIGNEE_USERNAME = "assigneeUsername"
    SUBJECT_USERNAME = "subjectUsername"


class CaseStatus(_Enum):
    """Possible statuses for a case."""

    CLOSED = "CLOSED"
    OPEN = "OPEN"
    ARCHIVED = "ARCHIVED"


class FileAvailability(_Enum):
    """Possible states of file-event source file availability."""

    error = "ERROR"
    exact_file_available = "EXACT_FILE_AVAILABLE"
    no_file_available = "NO_FILE_AVAILABLE"
    pending = "PENDING"
    recent_file_available = "RECENT_FILE_AVAILABLE"



import click

start_option = click.option(
    "--start",
    default=None,
    help="The beginning of the date range in which to look for audit log events. Accepts a date/time in yyyy-MM-dd (UTC) or"
    "yyyy-MM-dd HH:MM:SS (UTC+24-hr time) format.",
)
end_option = click.option(
    "--end",
    default=None,
    help="The end of the date range in which to look for audit log events, argument format options are the same as `--start`.",
)
actor_ids_option = click.option(
    "--actor-ids",
    default=None,
    help="Filter events by the actor IDs. Comma-delimited string of actor user IDs.",
)
actor_ip_addresses_option = click.option(
    "--actor-ip-addresses",
    default=None,
    help="Filter events by actor IP addresses. Comma-delimited string of actor IP addresses.",
)
actor_names_option = click.option(
    "--actor-names",
    default=None,
    help="Filter events by actor usernames. Comma-delimited string of actor usernames.",
)
event_types_option = click.option(
    "--event-types",
    default=None,
    help="Filter events by event type. Comma delimited string of event types.",
)
resource_ids_option = click.option(
    "--resource-ids",
    default=None,
    help="Filter events by resource ID. Comma delimited string of resource IDs.",
)
user_types_option = click.option(
    "--user-types",
    default=None,
    help="Filter events by user type. Comma delimited string of user types.",
)


def filter_options(f):
    f = start_option(f)
    f = end_option(f)
    f = actor_ids_option(f)
    f = actor_ip_addresses_option(f)
    f = actor_names_option(f)
    f = event_types_option(f)
    f = resource_ids_option(f)
    f = user_types_option(f)
    return f
